import pandas as pd
import os
from scipy.stats import pearsonr


def read_sol(dir_sol):
    sol = pd.read_csv(dir_sol, header=None)[0].str.split(expand=True)
    sol.columns = sol.columns + 1
    return sol


def get_breeding_effects(dir_sol):
    sol = read_sol(dir_sol)
    breeding_effects = sol.loc[sol[1] == "4", :].loc[:, [5, 8]]
    breeding_effects = dict(
        zip(breeding_effects.iloc[:, 0], breeding_effects.iloc[:, 1])
    )
    # 将key转换为int
    breeding_effects = {int(k): float(v) for k, v in breeding_effects.items()}
    return breeding_effects


# 计算遗传力相关
def cal_bv_cor(predict: dict, data: pd.DataFrame, pedigree: pd.DataFrame):
    """计算遗传力相关系数
    Args:
        predict (dict): 预测的个体与表型的对应字典
        data (pd.DataFrame): 原始数据，用来跑动物模型，最后一行为表型
        pedigree (pd.DataFrame): 家系

        Returns:
            float: 相关系数
    """
    # 生成替换数据
    pre_data = get_replace_data(predict, data)
    # 运行
    try:
        pre_bv = cal_bv_ani(predict.keys(), pre_data, pedigree)
        ori_bv = cal_bv_ani(predict.keys(), data, pedigree)
    except:
        print("动物模型运行失败")
        return 0
    # 计算相关系数
    bv_cor = cal_cor(pre_bv, ori_bv)
    print("遗传力相关系数：", bv_cor)
    clean_dmu_files()
    return bv_cor


# 计算相关系数
def cal_cor(pre_bv: dict, ori_bv: dict) -> float:
    # 选取共同的ID
    ids = set(pre_bv.keys()) & set(ori_bv.keys())
    # 选取共同的BV
    pre_bv = {k: pre_bv[k] for k in ids}
    ori_bv = {k: ori_bv[k] for k in ids}
    # 计算相关系数
    bv_cor = pearsonr(list(pre_bv.values()), list(ori_bv.values()))[0]
    return bv_cor


# 生成替换数据
def get_replace_data(predict: dict, data):
    # 生成替换数据
    data = data.copy()
    # 表型，最后一列
    pheno = data.iloc[:, -1].copy()
    # 替换表型
    data["predict"] = data["id"].map(predict)
    data["predict"] = data["predict"].fillna(pheno)
    # 删除phenotype列
    data = data.drop(columns=[pheno.name])
    return data


# 动物模型计算遗传力
def cal_bv_ani(target_ids: list, data, pedigree) -> dict:
    # 运行
    run_ani(data, pedigree)
    # 读取结果
    breeding_effects = get_breeding_effects(write_ani_dir() + ".sol")
    # 选取目标
    bv = {k: breeding_effects[k] for k in target_ids}
    return bv


# 跑动物模型
def run_ani(data: pd.DataFrame, pedigree: pd.DataFrame):
    data.to_csv("data.txt", sep=" ", index=False, header=False)
    pedigree.to_csv("ped.txt", sep=" ", index=False, header=False)
    os.system("run_dmuai.bat " + write_ani_dir())


# 测定日模型计算遗传力
def cal_bv_tsd(target_ids: list, data, pedigree) -> pd.DataFrame:
    pass


# 跑测定日模型
def run_tsd(data, pedigree):
    data.to_csv("data.txt", sep=" ", index=False, header=False)
    pedigree.to_csv("ped.txt", sep=" ", index=False, header=False)
    os.system("run_dmuai.bat TSD")


# 删除dmu生成文件
def clean_dmu_files(excepts=[]):
    # 根据后缀删除
    d_drop = [".SOL", ".log", ".LLIK", ".lst", ".PAROUT", ".RESIDUAL", ".PAROUT_STD", "data.txt", "ped.txt"]
    for except_ in excepts:
        d_drop.remove(except_)
    for root, dirs, files in os.walk("./"):
        for file in files:
            for d in d_drop:
                if file.endswith(d):
                    os.remove(os.path.join(root, file))


# 生成动物模型的DIR文件
def write_ani_dir(DIR="AAAAAAA"):
    try:
        os.system.remove(DIR + ".DIR")
    except:
        pass
    COMMENT = "\n"
    ANALYSE = "1 1 0 1"
    DATA = "ASCII (3,1,-999) ./data.txt"
    VARIABLE_1 = "#1    2       3"
    VARIABLE_2 = "BATCH STATION ID"
    VARIABLE_3 = "#1"
    VARIABLE_4 = "PHENO"
    MODEL_1 = "1 1 0 0 0"
    MODEL_2 = "0"
    MODEL_3 = "1 0 3 1 2 3"
    MODEL_4 = "# 1个随机效应（遗传效应）"
    model_5 = "1"
    model_6 = "0"
    model_7 = "0"
    VAR_STR = "1 PED 2 ASCII ./ped.txt"
    RESIDUALS = "ASCII"
    DMUAI_1 = "10"
    DMUAI_2 = "1D-7"
    DMUAI_3 = "1D-6"
    DMUAI_4 = "1"
    DMUAI_5 = "0"
    with open(DIR + ".DIR", "w") as f:
        f.write("$COMMENT\n")
        f.write(COMMENT + "\n")
        f.write("$ANALYSE ")
        f.write(ANALYSE + "\n")
        f.write("$DATA ")
        f.write(DATA + "\n")
        f.write("$VARIABLE\n")
        f.write(VARIABLE_1 + "\n")
        f.write(VARIABLE_2 + "\n")
        f.write(VARIABLE_3 + "\n")
        f.write(VARIABLE_4 + "\n")
        f.write("$MODEL\n")
        f.write(MODEL_1 + "\n")
        f.write(MODEL_2 + "\n")
        f.write(MODEL_3 + "\n")
        f.write(MODEL_4 + "\n")
        f.write(model_5 + "\n")
        f.write(model_6 + "\n")
        f.write(model_7 + "\n")
        f.write("$VAR_STR ")
        f.write(VAR_STR + "\n")
        f.write("$RESIDUALS ")
        f.write(RESIDUALS + "\n")
        f.write("$DMUAI\n")
        f.write(DMUAI_1 + "\n")
        f.write(DMUAI_2 + "\n")
        f.write(DMUAI_3 + "\n")
        f.write(DMUAI_4 + "\n")
        f.write(DMUAI_5 + "\n")
    return DIR

# 计算遗传力
def cal_bv(dir_lst:str ) -> float:
    # 读取dir文件
    with open(dir_lst, "r") as f:
        # 出现 FINAL PARAMETERS ESTIMATED 之后
        # 出现 Covariance matrix for random effect no:            1
        pass


# 测定日

import pandas as pd
import numpy as np
import os
from scipy.stats import pearsonr

class testday_sol_reader:

    def __init__(self, dir_sol, dir_data, id_index) -> None:
        """
        :param dir_sol: sol文件路径
        :param dir_data: data文件路径
        :param id_index: 个体号在data中在第几列
        """

        self.dir_sol = dir_sol
        self.dir_data = dir_data
        self.id_index = id_index

        self.sol = self.read_sol()
        self.data = self.read_data()
        self.effect_values = []

    def read_sol(self):
        sol = pd.read_csv(self.dir_sol, header=None)
        # 使用str.split()函数分割数据并扩展成多列
        sol[["1", "2", "阶数", "4", "个体号", "6", "7", "8", "9"]] = sol[0].str.split(
            expand=True
        )
        sol = sol.drop(0, axis=1)
        sol = sol[sol["1"] == "4"]
        return sol

    def read_data(self):
        data = pd.read_csv(self.dir_data, header=None, sep=" ")
        return data

    def calculate_ebv(self):

        # 每个阶数上每个个体的sol列
        effect_values = {}
        for degree, group in self.sol.groupby("阶数"):
            # 获取该阶数的sol列值并创建DataFrame
            effect_value = group[["个体号", "8"]]
            # 将DataFrame添加到字典中
            effect_values[degree] = effect_value


        ebv = {}
        for id in self.data[self.id_index].unique():
            ebv[id] = 0

        for degree in range(1, self.sol["阶数"].astype(int).max() + 1):
            print(degree)
            # 此阶数所有个体各个时间上的log值
            log_value = {}
            for id, group in self.data.groupby(by=self.id_index):
                log_value[id] = list(group[self.id_index + degree])
            # 累加EBV
            for id in self.data[self.id_index].unique():
                effect_value = effect_values[str(degree)]
                effect_value = float(effect_value[effect_value["个体号"] == str(id)]["8"])
                ebv[id] += (pd.Series(log_value[id]) * effect_value).mean()
                # ebv[id] += effect_value
        return ebv


import pandas as pd
import numpy as np
from multiprocessing import Pool

# 用第一种方法筛选采食数据，并记录每种错误的数量
def get_feed_data_filter_1(raw_data):
    # 将数据按照id——enter_time排序
    raw_data = raw_data.sort_values(by=['id', 'enter_time'])
    # 删除每个id的第一条和最后一条数据
    raw_data = raw_data.groupby('id').apply(
        lambda x: x.iloc[1:-1]).reset_index(drop=True)
    # 存储各种错误的数量
    error_num = {}
    error_index = []
    # FIV-low:每次采食量<-20g的数据
    error_num['FIV-low'] = len(
        raw_data[raw_data['feed_amount'] < -20])
    error_index.extend(raw_data[raw_data['feed_amount'] < -20].index)
    # FIV-high:每次采食量>2000g的数据
    error_num['FIV-high'] = len(
        raw_data[raw_data['feed_amount'] > 2000])
    error_index.extend(raw_data[raw_data['feed_amount'] > 2000].index)
    # FIV-0:每次采食量=0的数据
    error_num['FIV-0'] = len(raw_data[raw_data['feed_amount'] == 0])
    error_index.extend(raw_data[raw_data['feed_amount'] == 0].index)
    # OTV-low:每次采食时间<0的数据
    error_num['OTV-low'] = len(
        raw_data[raw_data['feed_time'] < 0])
    error_index.extend(raw_data[raw_data['feed_time'] < 0].index)
    # OTV-high:每次采食时间>3600s的数据
    error_num['OTV-high'] = len(
        raw_data[raw_data['feed_time'] > 3600])
    error_index.extend(raw_data[raw_data['feed_time'] > 3600].index)
    # FRV-High-FIV-low:每次采食率>500g/min且 0g<每次采食量<50g的数据
    error_num['FRV-High-FIV-low'] = len(raw_data[((raw_data['feed_amount'] > 0) & (
        raw_data['feed_amount'] < 50) & (raw_data['feed_time'] < 6))])
    error_index.extend(raw_data[((raw_data['feed_amount'] > 0) & (
        raw_data['feed_amount'] < 50) & (raw_data['feed_time'] < 6))].index)
    # FRV-high-strict:每次采食率>100g/min & 每次采食量≥50g或者每次采食量＜-20g
    error_num['FRV-high-strict'] = len(raw_data[((raw_data['feed_amount'] >= 50) | (
        raw_data['feed_amount'] < -20)) & (raw_data['feed_time'] < (50/(100/60)))])
    error_index.extend(raw_data[((raw_data['feed_amount'] >= 50) | (
        raw_data['feed_amount'] < -20)) & (raw_data['feed_time'] < (50/(100/60)))].index)
    # FRV-high:每次采食率>170g/min & 每次采食量≥50g或者每次采食量＜-20g的数据
    error_num['FRV-high'] = len(raw_data[((raw_data['feed_amount'] >= 50) | (
        raw_data['feed_amount'] < -20)) & (raw_data['feed_time'] < (50/(170/60)))])
    error_index.extend(raw_data[((raw_data['feed_amount'] >= 50) | (
        raw_data['feed_amount'] < -20)) & (raw_data['feed_time'] < (50/(170/60)))].index)
    # FRV-0:每次采食率=0g/min & 采食时间>500s
    error_num['FRV-0'] = len(raw_data[((raw_data['feed_amount']
                                        == 0) & (raw_data['feed_time'] > 500))])
    error_index.extend(raw_data[((raw_data['feed_amount']
                                == 0) & (raw_data['feed_time'] > 500))].index)
    # FRV-low:每次采食率!=0g/min & 采食率<2g/min
    error_num['FRV-low'] = len(raw_data[((raw_data['feed_amount'] != 0) & (
        (raw_data['feed_amount'] / raw_data['feed_time'])*60 < 2))])
    error_index.extend(raw_data[((raw_data['feed_amount'] != 0) & (
        (raw_data['feed_amount'] / raw_data['feed_time'])*60 < 2))].index)
    # all:所有错误的数据
    error_num['all'] = len(raw_data[((raw_data['feed_amount'] < -20) | (raw_data['feed_amount'] > 2000) | (raw_data['feed_amount'] == 0) | (raw_data['feed_time'] < 0) | (raw_data['feed_time'] > 3600) | ((raw_data['feed_amount'] > 0) & (raw_data['feed_amount'] < 50) & (raw_data['feed_time'] < 6))
                                    | ((raw_data['feed_amount'] >= 50) & (raw_data['feed_time'] < 6)) | ((raw_data['feed_amount'] >= 50) & (raw_data['feed_time'] < 6)) | ((raw_data['feed_amount'] == 0) & (raw_data['feed_time'] > 500)) | ((raw_data['feed_amount'] != 0) & (raw_data['feed_time'] < 2)))])
    # 返回筛选后的数据
    return raw_data.drop(error_index)

# 用第二种方法筛选采食数据，并记录每种错误的数量
def get_feed_data_filter_2(raw_data):
    # 需要的数据：id,enter_time,exit_time,feed_amount,feed_time,weight,birth_date
    # 单位：feed_amount:g,feed_time:s

    # 计算FRV = feed rate value (feed amount / feed time)，g/min
    raw_data['FRV'] = raw_data['feed_amount'] / raw_data['feed_time'] * 60
    # 计算FRV的上下限
    FRV_up = raw_data['FRV'].quantile(0.75) + \
        1.5*(raw_data['FRV'].quantile(0.75) -
             raw_data['FRV'].quantile(0.25))
    FRV_down = raw_data['FRV'].quantile(0.25) - \
        1.5*(raw_data['FRV'].quantile(0.75) -
             raw_data['FRV'].quantile(0.25))
    # 计算OTV的上下限
    OTV_up = raw_data['feed_time'].quantile(0.75) + \
        1.5*(raw_data['feed_time'].quantile(0.75) -
             raw_data['feed_time'].quantile(0.25))
    OTV_down = raw_data['feed_time'].quantile(0.25) - \
        1.5*(raw_data['feed_time'].quantile(0.75) -
             raw_data['feed_time'].quantile(0.25))
    # 计算前向体重差 LWD = leading weight difference (weight of following visit − weight of present visit)
    raw_data['LWD'] = raw_data.groupby(
        'id')['weight'].shift(-1) - raw_data['weight']
    # 计算LWD的上下限
    LWD_up = raw_data['LWD'].quantile(0.75) + \
        1.5*(raw_data['LWD'].quantile(0.75) -
             raw_data['LWD'].quantile(0.25))
    LWD_down = raw_data['LWD'].quantile(0.25) - \
        1.5*(raw_data['LWD'].quantile(0.75) -
             raw_data['LWD'].quantile(0.25))
    # 计算后向体重差 FWD = following weight difference (weight of present visit − weight of preceding visit)
    raw_data['FWD'] = raw_data['weight'] - \
        raw_data.groupby('id')['weight'].shift(1)
    # 计算FWD的上下限
    FWD_up = raw_data['FWD'].quantile(0.75) + \
        1.5*(raw_data['FWD'].quantile(0.75) -
             raw_data['FWD'].quantile(0.25))
    FWD_down = raw_data['FWD'].quantile(0.25) - \
        1.5*(raw_data['FWD'].quantile(0.75) -
             raw_data['FWD'].quantile(0.25))
    # 计算前向时间差 LTD = leading time difference (entrance time of following visit − exit time of present visit)
    raw_data['LTD'] = raw_data.groupby(
        'id')['enter_time'].shift(-1) - raw_data['exit_time']
    raw_data['LTD'] = raw_data['LTD'].apply(lambda x: x.total_seconds())
    # 计算LTD的上下限
    LTD_up = raw_data['LTD'].quantile(0.75) + \
        1.5*(raw_data['LTD'].quantile(0.75) -
             raw_data['LTD'].quantile(0.25))
    LTD_down = raw_data['LTD'].quantile(0.25) - \
        1.5*(raw_data['LTD'].quantile(0.75) -
             raw_data['LTD'].quantile(0.25))
    # 计算后向时间差 FTD = following time difference (entrance time of present visit − exit time of preceding visit)
    raw_data['FTD'] = raw_data['enter_time'] - \
        raw_data.groupby('id')['exit_time'].shift(1)
    raw_data['FTD'] = raw_data['FTD'].apply(lambda x: x.total_seconds())
    # 计算FTD的上下限
    FTD_up = raw_data['FTD'].quantile(0.75) + \
        1.5*(raw_data['FTD'].quantile(0.75) -
             raw_data['FTD'].quantile(0.25))
    FTD_down = raw_data['FTD'].quantile(0.25) - \
        1.5*(raw_data['FTD'].quantile(0.75) -
             raw_data['FTD'].quantile(0.25))
    

    error_num = {}
    error_index = []
    # error1: 所有进出站数据中，每次采食量FIV<-20g的数据
    error1 = raw_data[raw_data['feed_amount'] < -20]
    error_index.extend(error1.index)
    error_num['error1'] = len(error1)
    # error2: 所有进出站数据中，每次采食量FIV>2000g的数据
    error2 = raw_data[raw_data['feed_amount'] > 2000]
    error_index.extend(error2.index)
    error_num['error2'] = len(error2)
    # error3: 采食时间OTV=0的数据中，|FIV|>20g的数据
    error3 = raw_data[(raw_data['feed_time'] == 0) &
                      (abs(raw_data['feed_amount']) > 20)]
    error_index.extend(error3.index)
    error_num['error3'] = len(error3)
    # error4: 所有进出站数据中，采食时间OTV<0s的数据
    error4 = raw_data[raw_data['feed_time'] < 0]
    error_index.extend(error4.index)
    error_num['error4'] = len(error4)
    # error5: 所有进出站数据中，采食时间OTV>OTV_up的数据
    error5 = raw_data[raw_data['feed_time'] > OTV_up]
    error_index.extend(error5.index)
    error_num['error5'] = len(error5)
    # error6: 0g<FIV<50g的数据中，采食速率FRV>500g/min的数据
    error6 = raw_data[(raw_data['feed_amount'] > 0) & (
        raw_data['feed_amount'] < 50) & (raw_data['FRV'] > 500)]
    error_index.extend(error6.index)
    error_num['error6'] = len(error6)
    # error7: FIV>50g的数据中，采食速率FRV>110g/min的数据
    error7 = raw_data[(raw_data['feed_amount'] > 50) & (
        raw_data['FRV'] > 110)]
    error_index.extend(error7.index)
    error_num['error7'] = len(error7)
    # error8: FIV>50g的数据中，采食速率FRV>FRV_up的数据
    error8 = raw_data[(raw_data['feed_amount'] > 50) & (
        raw_data['FRV'] > FRV_up)]
    error_index.extend(error8.index)
    error_num['error8'] = len(error8)
    # error9: FRV=0g/min的数据中，OTV>OTV_up的数据
    error9 = raw_data[(raw_data['feed_amount'] == 0)
                      & (raw_data['feed_time'] > OTV_up)]
    error_index.extend(error9.index)
    error_num['error9'] = len(error9)
    # error10: FRV!=0g/min的数据中，|FRV|<2g/min的数据
    error10 = raw_data[(raw_data['feed_amount'] != 0)
                       & (abs(raw_data['FRV']) < 2)]
    error_index.extend(error10.index)
    error_num['error10'] = len(error10)
    # error11: LWD<LWD_down的数据
    error11 = raw_data[raw_data['LWD'] < LWD_down]
    error_index.extend(error11.index)
    error_num['error11'] = len(error11)
    # error12: LWD>LWD_up的数据
    error12 = raw_data[raw_data['LWD'] > LWD_up]
    error_index.extend(error12.index)
    error_num['error12'] = len(error12)
    # error13: FWD<FWD_down的数据
    error13 = raw_data[raw_data['FWD'] < FWD_down]
    error_index.extend(error13.index)
    error_num['error13'] = len(error13)
    # error14: FWD>FWD_up的数据
    error14 = raw_data[raw_data['FWD'] > FWD_up]
    error_index.extend(error14.index)
    error_num['error14'] = len(error14)
    # error15: LTD<0s的数据
    error15 = raw_data[raw_data['LTD'] < 0]
    error_index.extend(error15.index)
    error_num['error15'] = len(error15)
    # error16: FTD<0s的数据
    error16 = raw_data[raw_data['FTD'] < 0]
    error_index.extend(error16.index)
    error_num['error16'] = len(error16)
    result = raw_data.drop(error_index)
    # all:所有错误的数据
    error_num['all'] = len(raw_data) - len(result)
    return result

# 就不筛了。按原来的步骤！
def get_feed_data_filter_3(raw_data):
    # 需要的数据：id,enter_time,exit_time,feed_amount,feed_time,weight,birth_date
    # 单位：feed_amount:g,feed_time:s

    # 计算FRV = feed rate value (feed amount / feed time)，g/min
    raw_data['FRV'] = raw_data['feed_amount'] / raw_data['feed_time'] * 60

    # 计算前向体重差 LWD = leading weight difference (weight of following visit − weight of present visit)
    raw_data['LWD'] = raw_data.groupby(
        'id')['weight'].shift(-1) - raw_data['weight']

    # 计算后向体重差 FWD = following weight difference (weight of present visit − weight of preceding visit)
    raw_data['FWD'] = raw_data['weight'] - \
        raw_data.groupby('id')['weight'].shift(1)

    # 计算前向时间差 LTD = leading time difference (entrance time of following visit − exit time of present visit)
    raw_data['LTD'] = raw_data.groupby(
        'id')['enter_time'].shift(-1) - raw_data['exit_time']
    raw_data['LTD'] = raw_data['LTD'].apply(lambda x: x.total_seconds())

    # 计算后向时间差 FTD = following time difference (entrance time of present visit − exit time of preceding visit)
    raw_data['FTD'] = raw_data['enter_time'] - \
        raw_data.groupby('id')['exit_time'].shift(1)
    raw_data['FTD'] = raw_data['FTD'].apply(lambda x: x.total_seconds())

    result = raw_data
    return result

def weight_filter(raw_data: pd.DataFrame):
    # 按id分组
    groups = raw_data.groupby('id')
    # 多进程筛选
    result = paralleize_dataframe(groups, apply_filt_group)
    return result

def filt_weight(group):
    """
    用于筛选体重数据
    n_std: 残差的标准差的倍数
    """
    group = group.copy()
    group.sort_values(by=['age'], inplace=True)
    # start_weight和weight不是一个单位欸
    group.at[group.index[0], 'weight'] = group.at[group.index[0], 'start_weight']
    group.at[group.index[-1], 'weight'] = group.at[group.index[-1], 'end_weight']
    # 只用开始和结束体重用于构建线性模型
    model = np.polyfit(group['age'].iloc[[0, -1]], group['weight'].iloc[[0, -1]], 1)
    # 计算残差
    group['residual'] = group['weight'] - (model[0]*group['age'] + model[1])

    # 计算残差的标准差
    std = group['residual'].std()
    # 将weight中超过残差绝对值三个标准差的数据改为nan
    group.loc[abs(group['residual']) > 3*std, 'weight'] = np.nan
    # 增加一列，记录这处是否改成了nan
    group['weight_is_nan'] = group['weight'].isna()
    # 增加一列，记录std
    group['std'] = std
    # 用移动平均法填充
    group['weight'] = group['weight'].fillna(
        group['weight'].rolling(5, min_periods=1).mean())
    return group

def apply_filt_group(group):
    return filt_weight(group)

def paralleize_dataframe(df_grouped, func):
    df_split = [group for name, group in df_grouped]
    pool = Pool(16)
    df = pd.concat(pool.map(func, df_split))
    pool.close()
    pool.join()
    df.drop(columns=['weight_is_nan', 'residual', 'std'], inplace=True)
    return df

def get_data(raw_data, filter_method=2):
    # 需要的数据：id,enter_time,exit_time,feed_amount,feed_time,weight,birth_date
    # 单位：feed_amount:g,feed_time:s
    uneed_columns = ['FTD', 'LTD', 'FWD', 'LWD', 'FRV', 'exit_time', 'feed_time']
    data = raw_data.copy()
    data['enter_time'] = pd.to_datetime(data['enter_time'])
    data['exit_time'] = pd.to_datetime(data['exit_time'])
    data['birth_date'] = pd.to_datetime(data['birth_date'])

    data['age'] = data['enter_time'] - data['birth_date']
    data['age'] = data['age'].dt.total_seconds()
    
    data = weight_filter(data)

    data['age'] = data['enter_time'] - data['birth_date']
    data['age'] = data['age'].dt.days

    final_data = data.copy()

    # 筛选采食数据
    if filter_method == 1:
        data = get_feed_data_filter_1(data)
    elif filter_method == 2:
        data = get_feed_data_filter_2(data)
    elif filter_method == 3:
        data = get_feed_data_filter_3(data)

    # 按id和date排序
    final_data = data.sort_values(by=['id', 'age'])

    # 将体重的零值填充为前后5次体重的平均值
    final_data['weight'] = final_data.groupby('id', group_keys=False)['weight'].apply(
        lambda x: x.replace(0, np.nan).fillna(method='ffill').fillna(method='bfill'))
    
    # 如果数据所有体重均值 > 1000g，那么将体重单位转换为kg
    if final_data['weight'].mean() > 1000:
        final_data['weight'] = final_data['weight'] / 1000
    if final_data['start_weight'].mean() > 1000:
        final_data['start_weight'] = final_data['start_weight'] / 1000
    if final_data['end_weight'].mean() > 1000:
        final_data['end_weight'] = final_data['end_weight'] / 1000

    # 如果数据所有采食均值 > 100g，那么将采食单位转换为kg
    if final_data['feed_amount'].mean() > 10:
        final_data['feed_amount'] = final_data['feed_amount'] / 1000

    # 如果有uneed_columns中的列，那么删除
    for column in uneed_columns:
        if column in final_data.columns:
            final_data = final_data.drop(columns=[column])

    # 改最小enter_time
    final_data['enter_time'] = final_data.groupby('id')['enter_time'].transform('min')
    # 求出每天的采食总量
    final_data['feed_amount'] = final_data.groupby(['id', 'age'])['feed_amount'].transform('sum')
    # 求出每天的平均体重
    final_data['weight'] = final_data.groupby(['id', 'age'])['weight'].transform('mean')

    # 删除重复的数据
    final_data.drop_duplicates(inplace=True)

    # 将第一条体重改成start_weight
    final_data.loc[final_data.groupby('id').head(1).index, 'weight'] = final_data['start_weight']
    # 将最后一条体重改成end_weight
    final_data.loc[final_data.groupby('id').tail(1).index, 'weight'] = final_data['end_weight']

    final_data.reset_index(drop=True, inplace=True)

    return final_data


import numpy as np
import pandas as pd
from numpy.polynomial import legendre
from scipy import stats
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder
import time

# 数据处理

class Data_Generator:
    def __init__(self, daily_data):

        self.raw_data = daily_data.copy()
        self.daily_data = self.get_daily_data()
        print("daily_data shape: ", self.daily_data.shape)
        self.id_encoder = self.get_id_encoder()
        print("id_encoder shape: ", self.id_encoder.classes_.shape)
        self.daily_data["id"] = self.id_encoder.transform(self.daily_data["id"]) + 1
        self.daily_data["father"] = (
            self.id_encoder.transform(self.daily_data["father"]) + 1
        )
        self.daily_data["mother"] = (
            self.id_encoder.transform(self.daily_data["mother"]) + 1
        )
        self.max_day = self.daily_data["age"].max()
        self.min_day = self.daily_data["age"].min()

    # 获得daily_data
    def get_daily_data(self):
        raw_data = self.raw_data.copy()
        # 排序
        raw_data = raw_data.sort_values(["id", "age"]).reset_index(drop=True)
        # 计算每天的增重
        raw_data["gain_weight"] = raw_data.groupby("id")["weight"].diff()
        return raw_data.reset_index(drop=True)

    # 对daily_data中每个个体的长度进行统计
    def get_id_length(self):
        id_length = self.daily_data.groupby("id").count()["age"]
        return id_length

    # 获得id_encoder
    def get_id_encoder(self):
        ids = pd.concat(
            [
                self.daily_data["id"],
                self.daily_data["father"],
                self.daily_data["mother"],
            ]
        ).unique()
        id_encoder = LabelEncoder()
        id_encoder.fit(ids)
        return id_encoder

    # 根据dataframe中的“age”，获得每条数据对应lg值
    def get_lg_df(self, input_data, degree=5):
        lg_df = input_data.copy()[["age", "id"]]
        min_day = lg_df["age"].min()
        max_day = lg_df["age"].max()
        lg_df["standared_day"] = 2 * (lg_df["age"] - min_day) / (max_day - min_day) - 1
    
        for i in range(degree + 1):
            lg_df['lg' + str(i)] = legendre.legval(lg_df['standared_day'], [0] * i + [1]) * np.sqrt((2 * i + 1) / 2)
        
        return lg_df.drop(["standared_day", "age", "id"], axis=1)

    # 根据输入数据，获得batch_df: 年-季-年龄
    def get_batch_df(self, input_data):
        batch_df = input_data[["id"]].copy()

        # 按个体号，进入时间排序
        input_data = input_data.sort_values(["id", "age"]).reset_index(drop=True).copy()
        data = (
            input_data.groupby("id")
            .first()
            .reset_index(drop=False)[["id", "enter_time", "age"]]
            .copy()
        )
        data["year"] = data["enter_time"].apply(lambda t: t.year)
        data["season"] = data["enter_time"].apply(lambda t: t.quarter)
        data["age"] = data["age"].apply(lambda t: int(t / 10))

        # 结合年份和季度，获得year-season
        data["batch"] = (
            data["year"].astype(str)
            + "-"
            + data["season"].astype(str)
            + "-"
            + data["age"].astype(str)
        )

        batch_df = batch_df.merge(data[["id", "batch"]], on="id", how="left")

        return batch_df[["batch"]]
    

    def get_dmu_data(
        self, sep=5, if_max_breed=False, if_max_sex=False, min_test_length=0
    ):
        target_columns = [
            "sex",
            "batch",
            "breed",
            "station",
            "id",
            "lg0",
            "lg1",
            "lg2",
            "lg3",
            "lg4",
            "lg5",
            "dfi",
            "dgw",
            "fcr",
        ]
        final_data = self.daily_data.copy()

        start_time = time.time()
        # 删除数据条数小于min_test_length的个体
        id_length = self.get_id_length()
        id_length = id_length[id_length >= min_test_length]
        final_data = final_data[final_data["id"].isin(id_length.index)]
        print("删除数据条数小于min_test_length的个体: ", time.time() - start_time)
        start_time = time.time()
        # 删除每个id的第一条和最后一条数据
        final_data = (
            final_data.groupby("id")
            .apply(lambda x: x.iloc[1:-1])
            .reset_index(drop=True)
        )
        print("删除每个id的第一条和最后一条数据: ", time.time() - start_time)

        feed_data = pd.DataFrame()
        feed_datas = []
        start_time = time.time()
        for id, data in final_data.groupby("id"):
            bins = int(len(data) / sep)
            groups = pd.cut(data.index, bins=bins, labels=False)
            data["group"] = groups

            new_data = data.groupby("group")[["feed_amount", "gain_weight"]].mean(
                numeric_only=False
            )

            new_data["id"] = id
            new_data["age"] = data.groupby("group").last()["age"]
            new_data["fcr"] = new_data["feed_amount"] / new_data["gain_weight"]

            feed_datas.append(new_data)
            
        feed_data = pd.concat(feed_datas, axis=0)
        
        feed_data.rename(
            columns={"feed_amount": "dfi", "gain_weight": "dgw"}, inplace=True
        )
        print("feed_data: ", time.time() - start_time)

        start_time = time.time()

        final_data = final_data.drop(["feed_amount", "gain_weight", "weight"], axis=1)
        final_data = feed_data.merge(final_data, on=["id", "age"])
        print("merge: ", time.time() - start_time)

        start_time = time.time()
        # 获得lg_df
        lg_df = self.get_lg_df(final_data)
        print("lg_df: ", time.time() - start_time)
        start_time = time.time()
        # 获得ys_df
        ys_df = self.get_batch_df(final_data)
        print("ys_df: ", time.time() - start_time)
        start_time = time.time()
        # 左右拼接数据
        final_data = pd.concat([final_data, lg_df, ys_df], axis=1)
        print("concat: ", time.time() - start_time)
        start_time = time.time()
        # 对object类型的数据进行编码
        final_data = self.recode_object_data(final_data)
        print("recode: ", time.time() - start_time)
        start_time = time.time()
        xp = final_data[["id", "father", "mother"]].copy()
        xp["year"] = 2018
        xp = xp.drop_duplicates().reset_index(drop=True)
        print("xp: ", time.time() - start_time)

        start_time = time.time()
        # 是否删除其它品种
        if if_max_breed:
            # 选择数量最多的品种
            final_data = self.get_max_breed_data(final_data)
            # 在target_columns中删除breed
            target_columns.remove("breed")
        print("if_max_breed: ", time.time() - start_time)
        start_time = time.time()

        # 是否删除其它性别
        if if_max_sex:
            # 选择数量最多的性别
            final_data = self.get_max_sex_data(final_data)
            # 在target_columns中删除sex
            target_columns.remove("sex")
        print("if_max_sex: ", time.time() - start_time)
        return final_data[target_columns].round(4), xp

    def recode_object_data(self, data):
        for col in data.columns:
            if data[col].dtype == "object":
                data[col] = LabelEncoder().fit_transform(data[col]) + 1
        return data

    def get_rep_data(
        self, sep=5, if_max_breed=False, if_max_sex=False, min_test_length=0
    ):
        final_data, xp = self.get_dmu_data(
            sep, if_max_breed, if_max_sex, min_test_length
        )
        # 删除不需要的列
        final_data = final_data.drop(["lg0", "lg1", "lg2", "lg3", "lg4", "lg5"], axis=1)
        return final_data.round(2), xp

    def get_ani_data(self, if_max_breed=False, if_max_sex=False, min_test_length=0):
        final_data, xp = self.get_dmu_data(1, if_max_breed, if_max_sex, min_test_length)

        data1 = final_data[["id", "dfi", "dgw"]].copy()

        data1 = data1.groupby("id").sum()
        data1.reset_index(inplace=True, drop=False)
        data1["fcr"] = data1["dfi"] / data1["dgw"]

        data2 = (
            final_data.drop(
                ["lg0", "lg1", "lg2", "lg3", "lg4", "lg5", "dfi", "dgw", "fcr"], axis=1
            )
            .copy()
            .drop_duplicates()
        )

        data = data1.merge(data2, on="id", how="left")
        data = data[data2.columns.tolist() + ["dfi", "dgw", "fcr"]]
        return data.round(2), xp

    def get_normal_data(self):
        data = self.daily_data.copy()
        batch_df = self.get_batch_df(data)
        data = pd.concat([data, batch_df], axis=1)
        # 按照id，年龄排序
        data = data.sort_values(["id", "age"]).reset_index(drop=True)
        # 获得每个id的数据条数
        data["count"] = data.groupby("id")["id"].transform("count")
        # 计算每个id的日均增重
        data["gain_weight"] = data["end_weight"] - data["start_weight"]
        data["dgw"] = data["gain_weight"] / data["count"]
        # 计算每个id的日均采食
        data["total_feed"] = data.groupby("id")["feed_amount"].transform("sum") / 1000
        data["dfi"] = data["total_feed"] / data["count"]
        # 计算每个id的日均饲料转化率
        data["fcr"] = data["total_feed"] / data["gain_weight"]

        data = self.recode_object_data(data)

        return (
            data.drop(
                [
                    "gain_weight",
                    "total_feed",
                    "gain_weight",
                    "count",
                    "age",
                    "feed_amount",
                    "index",
                    "birth_date",
                    "weight",
                    "enter_time",
                    "birth_weight",
                    "start_weight",
                    "end_weight",
                ],
                axis=1,
            )
            .round(2)
            .drop_duplicates()
            .reset_index(drop=True)
        )

    def get_range_normal_data_1(
        self, start, end, sep=7, start_range=2, chunk_num=5, if_max_sex=False, if_max_breed=False, if_backwards=False
    ):
        ini_start = start
        ini_end = end
        data = self.daily_data.copy()
        batch_df = self.get_batch_df(data)
        data = pd.concat([data, batch_df], axis=1)

        # 按照id，年龄排序
        data = data.sort_values(["id", "age"]).reset_index(drop=True)

        wanted_id = self.get_id_by_age_range(ini_start, ini_end)

        target_datas = []

        if if_backwards:
            start = ini_end
            end = ini_end
        else:
            start = ini_start
            end = ini_start

        for _ in range(chunk_num):
            if if_backwards:
                start = start - sep - (start_range - 1) * sep
                end = end
            else:
                start = start
                end = end + sep + (start_range - 1) * sep
            # 删除不需要的id
            target_data = data[data["id"].isin(wanted_id)].copy()
            #  只保留日龄内数据
            target_data = target_data[
                (target_data["age"] >= ini_start) & (target_data["age"] <= ini_end)
            ]
            # 获得每个id的数据条数
            target_data["count"] = target_data.groupby("id")["id"].transform("count")
            # 将开测体重换成每个id的第一条数据的体重
            target_data["start_weight"] = target_data.groupby("id")["weight"].transform(
                "first"
            )
            # 将结测体重换成每个id的最后一条数据的体重
            target_data["end_weight"] = target_data.groupby("id")["weight"].transform(
                "last"
            )
            # 计算每个id的日均增重
            target_data["gain_weight"] = (
                target_data["end_weight"] - target_data["start_weight"]
            )
            target_data["dgw"] = target_data["gain_weight"] / target_data["count"]
            # 计算每个id的日均采食
            target_data["total_feed"] = target_data.groupby("id")[
                "feed_amount"
            ].transform("sum")
            target_data["dfi"] = target_data["total_feed"] / target_data["count"]
            # 计算每个id的日均饲料转化率
            target_data["fcr"] = target_data["total_feed"] / target_data["gain_weight"]

            # 计算时间范围内的数据
            range_data = target_data[
                (target_data["age"] >= start) & (target_data["age"] <= end)
            ].copy()
            # 获得每个id的数据条数
            range_data["count"] = range_data.groupby("id")["id"].transform("count")
            # 将开测体重换成每个id的第一条数据的体重
            range_data["start_weight"] = range_data.groupby("id")["weight"].transform(
                "first"
            )
            # 将结测体重换成每个id的最后一条数据的体重
            range_data["end_weight"] = range_data.groupby("id")["weight"].transform(
                "last"
            )
            # 计算每个id的日均增重
            range_data["gain_weight"] = (
                range_data["end_weight"] - range_data["start_weight"]
            )
            range_data["range_dgw"] = range_data["gain_weight"] / range_data["count"]
            # 计算每个id的日均采食
            range_data["total_feed"] = range_data.groupby("id")[
                "feed_amount"
            ].transform("sum")
            range_data["range_dfi"] = range_data["total_feed"] / range_data["count"]
            # 计算每个id的日均饲料转化率
            range_data["range_fcr"] = (
                range_data["total_feed"] / range_data["gain_weight"]
            )
            # 只保留id, range_dgw, range_dfi, range_fcr
            range_data = (
                range_data[["id", "range_dgw", "range_dfi", "range_fcr"]]
                .drop_duplicates()
                .reset_index(drop=True)
            )

            # 将range_data与data合并
            target_data = target_data.merge(range_data, on="id", how="left")

            target_data = self.recode_object_data(target_data)
            target_data = (
                target_data.drop(
                    [
                        "gain_weight",
                        "total_feed",
                        "gain_weight",
                        "count",
                        "age",
                        "feed_amount",
                        "birth_date",
                        "weight",
                        "enter_time",
                        "birth_weight",
                        "start_weight",
                        "end_weight",
                    ],
                    axis=1,
                )
                .round(2)
                .drop_duplicates()
                .reset_index(drop=True)
            )
            # 是否删除其它品种
            if if_max_breed:
                # 选择数量最多的品种
                target_data = self.get_max_breed_data(target_data)
                # 在target_columns中删除breed
                target_data = target_data.drop(["breed"], axis=1)
            # 是否删除其它性别
            if if_max_sex:
                target_data = self.get_max_sex_data(target_data)
                # 在target_columns中删除sex
                target_data = target_data.drop(["sex"], axis=1)
            target_datas.append(target_data)

        return target_datas
    
    def get_range_normal_data_2(
        self, start, end, sep=7, chunk_num=5, if_max_sex=False, if_max_breed=False, if_backwards=False
    ):
        ini_start = start
        ini_end = end
        data = self.daily_data.copy()
        batch_df = self.get_batch_df(data)
        data = pd.concat([data, batch_df], axis=1)

        # 按照id，年龄排序
        data = data.sort_values(["id", "age"]).reset_index(drop=True)

        target_datas = []

        for _ in range(chunk_num):

            if if_backwards:
                start = start - sep
                end = end
            else:
                start = start
                end = end + sep

            wanted_id = self.get_id_by_age_range(start, end)
            # 删除不需要的id
            target_data = data[data["id"].isin(wanted_id)].copy()

            #  只保留日龄内数据
            target_data = target_data[
                (target_data["age"] >= start) & (target_data["age"] <= end)
            ]
            # 获得每个id的数据条数
            target_data["count"] = target_data.groupby("id")["id"].transform("count")
            # 将开测体重换成每个id的第一条数据的体重
            target_data["start_weight"] = target_data.groupby("id")["weight"].transform(
                "first"
            )
            # 将结测体重换成每个id的最后一条数据的体重
            target_data["end_weight"] = target_data.groupby("id")["weight"].transform(
                "last"
            )
            # 计算每个id的日均增重
            target_data["gain_weight"] = (
                target_data["end_weight"] - target_data["start_weight"]
            )
            target_data["dgw"] = target_data["gain_weight"] / target_data["count"]
            # 计算每个id的日均采食
            target_data["total_feed"] = target_data.groupby("id")[
                "feed_amount"
            ].transform("sum")
            target_data["dfi"] = target_data["total_feed"] / target_data["count"]
            # 计算每个id的日均饲料转化率
            target_data["fcr"] = target_data["total_feed"] / target_data["gain_weight"]


            # 计算时间范围内的数据
            range_data = target_data[
                (target_data["age"] >= ini_start) & (target_data["age"] <= ini_end)
            ].copy()
            # 获得每个id的数据条数
            range_data["count"] = range_data.groupby("id")["id"].transform("count")
            # 将开测体重换成每个id的第一条数据的体重
            range_data["start_weight"] = range_data.groupby("id")["weight"].transform(
                "first"
            )
            # 将结测体重换成每个id的最后一条数据的体重
            range_data["end_weight"] = range_data.groupby("id")["weight"].transform(
                "last"
            )
            # 计算每个id的日均增重
            range_data["gain_weight"] = (
                range_data["end_weight"] - range_data["start_weight"]
            )
            range_data["range_dgw"] = range_data["gain_weight"] / range_data["count"]
            # 计算每个id的日均采食
            range_data["total_feed"] = range_data.groupby("id")[
                "feed_amount"
            ].transform("sum")
            range_data["range_dfi"] = range_data["total_feed"] / range_data["count"]
            # 计算每个id的日均饲料转化率
            range_data["range_fcr"] = (
                range_data["total_feed"] / range_data["gain_weight"]
            )
            # 只保留id, range_dgw, range_dfi, range_fcr
            range_data = (
                range_data[["id", "range_dgw", "range_dfi", "range_fcr"]]
                .drop_duplicates()
                .reset_index(drop=True)
            )

            # 将range_data与data合并
            target_data = target_data.merge(range_data, on="id")

            target_data = self.recode_object_data(target_data)
            target_data = (
                target_data.drop(
                    [
                        "gain_weight",
                        "total_feed",
                        "gain_weight",
                        "count",
                        "age",
                        "feed_amount",
                        "birth_date",
                        "weight",
                        "enter_time",
                        "birth_weight",
                        "start_weight",
                        "end_weight",
                    ],
                    axis=1,
                )
                .round(2)
                .drop_duplicates()
                .reset_index(drop=True)
            )
            # 是否删除其它品种
            if if_max_breed:
                # 选择数量最多的品种
                target_data = self.get_max_breed_data(target_data)
                # 在target_columns中删除breed
                target_data = target_data.drop(["breed"], axis=1)
            # 是否删除其它性别
            if if_max_sex:
                target_data = self.get_max_sex_data(target_data)
                # 在target_columns中删除sex
                target_data = target_data.drop(["sex"], axis=1)
            target_datas.append(target_data)


        return target_datas
    
    # 生成时间区段原始数据
    def get_chunk_data(self, start_age, end_age, sep):
        data = self.get_data_by_age_range(start_age, end_age)

        batch_df = self.get_batch_df(data)
        data = pd.concat([data, batch_df], axis=1)
        data["age"] = data["age"] - start_age

        data["chunk"] = data["age"].apply(lambda t: int(t / sep))
        data["feed"] = data.groupby("id")["feed_amount"].transform("sum")
        data["gain"] = data.groupby("id")["gain_weight"].transform("sum")
        data["fcr"] = data["feed"] / data["gain"]

        data["chunk_feed"] = data.groupby(["id", "chunk"])["feed_amount"].transform(
            "sum"
        )
        data["chunk_gain"] = data.groupby(["id", "chunk"])["gain_weight"].transform(
            "sum"
        )
        data["chunk_fcr"] = data["chunk_feed"] / data["chunk_gain"]

        # 将每个个体的chunk_fcr, chunk_feed, chunk_gain放到每个个体的同一条数据中
        chunk_datas = (
            data[["id", "chunk", "chunk_fcr", "chunk_feed", "chunk_gain"]]
            .drop_duplicates()
            .reset_index(drop=True)
        )

        chunk_datas["chunk_name_fcr"] = chunk_datas["chunk"].apply(
            lambda t: "chunk_fcr_" + str(t)
        )
        chunk_datas["chunk_name_feed"] = chunk_datas["chunk"].apply(
            lambda t: "chunk_feed_" + str(t)
        )
        chunk_datas["chunk_name_gain"] = chunk_datas["chunk"].apply(
            lambda t: "chunk_gain_" + str(t)
        )

        # print("start pivot")
        chunk_data = chunk_datas.pivot(
            index="id", columns="chunk_name_fcr", values="chunk_fcr"
        ).reset_index(drop=False)

        # 要哪些数据
        data = (
            data[
                [
                    "father",
                    "mother",
                    "birth_litter",
                    "litter_size",
                    "birth_weight",
                    "sex",
                    "batch",
                    "breed",
                    "station",
                    "id",
                    "fcr",
                ]
            ]
            .drop_duplicates()
            .reset_index(drop=True)
        )

        data = data.merge(chunk_data, on="id", how="left")

        data = self.recode_object_data(data)
        return data

    # 生成时间区段的普通数据
    def get_chunk_normal_data(
        self, start_age, end_age, sep, if_max_breed=False, if_max_sex=False
    ):
        data = self.get_chunk_data(start_age, end_age, sep)
        # if_max_breed, 获得数量最多的breed，将其它breed的数据删除
        if if_max_breed:
            data = self.get_max_breed_data(data)
            
            # 删除breed列
            data = data.drop(["breed"], axis=1)
        # if_max_sex, 获得数量最多的sex，将其它sex的数据删除
        if if_max_sex:
            data = self.get_max_sex_data(data)
            # 删除sex列
            data = data.drop(["sex"], axis=1)
        # 将为inf的数据替换为-999
        data = data.replace([np.inf, -np.inf], -999)
        return data

    # 通过开始与结束日龄，获得合适的id
    def get_id_by_age_range(self, start_age, end_age):
        data = self.daily_data.copy()
        data["start_age"] = data.groupby("id")["age"].transform("min")
        data["end_age"] = data.groupby("id")["age"].transform("max")
        # 中位数
        data["median_age"] = data.groupby("id")["age"].transform("median")

        return list(
            data[
                (data["start_age"] <= start_age) & (data["end_age"] >= end_age)
            ].id.unique()
        )

    # 描述一下daily_data的开始日龄与结束日龄分位数情况
    def describe_daily_data_age(self):
        data = self.daily_data.copy()
        data["start_age"] = data.groupby("id")["age"].transform("min")
        data["end_age"] = data.groupby("id")["age"].transform("max")
        data["median_age"] = data.groupby("id")["age"].transform("median")
        return data[["start_age", "end_age", "median_age"]].describe()

    # 描述一下各个周龄有多少个个体
    def describe_week_age(self):
        data = self.daily_data.copy()
        # 如果该周内数据不足7天，则删除该周
        data["week_age"] = data["age"] // 7 + 1
        data = data.groupby(["id", "week_age"]).filter(lambda x: len(x) >= 7)
        # 统计每个周龄有多少个个体
        data = data.groupby("week_age")["id"].nunique().reset_index(drop=False)
        data.columns = ["week_age", "count"]
        return data

    # 通过日龄范围，从daily_data中获取数据
    # 只保留日龄内的数据，其它的都删除
    def get_data_by_age_range(self, start_age, end_age):
        ids = self.get_id_by_age_range(start_age, end_age)
        data = self.daily_data.copy()
        data = data[data["id"].isin(ids)].copy()
        data = data[(data["age"] >= start_age) & (data["age"] <= end_age)]
        data.reset_index(drop=True, inplace=True)
        return data

    def get_pedigree_data(self):
        data = self.daily_data.copy()
        data = data[["id", "father", "mother"]].drop_duplicates()
        data["year"] = 2018
        data = data.drop_duplicates().reset_index(drop=True)
        return data

    def write_dmu_data(self, data, file_name):
        data.to_csv(file_name, sep=" ", index=False, header=False)

    def get_max_breed_data(self, data):
        data = data.copy()
        # 选择数量最多的品种
        breed_count = data.groupby("breed").count()["id"]
        max_breed = breed_count.idxmax()
        data = data[data["breed"] == max_breed]
        data.reset_index(drop=True, inplace=True)
        return data

    def get_max_sex_data(self, data):
        data = data.copy()
        # 选择数量最多的性别
        sex_count = data.groupby("sex").count()["id"]
        max_sex = sex_count.idxmax()
        data = data[data["sex"] == max_sex]
        data.reset_index(drop=True, inplace=True)
        return data